@@ -1,5 +1,7 @@
 import binascii
 import datetime as dt
+
+import jsonlines
 import logbook
 
 import attr
@@ -7,11 +9,14 @@ import attr
 from construct import *
 from io import BytesIO
 
+from tqdm import tqdm
+
 from regipy.exceptions import NoRegistrySubkeysException, RegistryKeyNotFoundException, NoRegistryValuesException, \
-    RegistryValueNotFoundException, RegipyGeneralException, UnidentifiedHiveException
+    RegistryValueNotFoundException, RegipyGeneralException, UnidentifiedHiveException, RegistryParsingException
+from regipy.hive_types import SUPPORTED_HIVE_TYPES
 from regipy.structs import REGF_HEADER, HBIN_HEADER, CM_KEY_NODE, LF_LH_SK_ELEMENT, VALUE_KEY, INDEX_ROOT, \
     REGF_HEADER_SIZE, INDEX_ROOT_SIGNATURE, LEAF_INDEX_SIGNATURE, FAST_LEAF_SIGNATURE, HASH_LEAF_SIGNATURE, \
-    BIG_DATA_BLOCK
+    BIG_DATA_BLOCK, INDEX_LEAF, DEFAULT_VALUE
 from regipy.utils import boomerang_stream, convert_wintime, identify_hive_type, MAX_LEN, try_decode_binary
 
 logger = logbook.Logger(__name__)
@@ -52,6 +57,9 @@ class Subkey:
     values_count = attr.ib(type=int)
     values = attr.ib(factory=list)
 
+    # This field will be used if a partial hive was given, if not it would be None.
+    actual_path = attr.ib(type=str, default=None)
+
 
 @attr.s
 class Value:
@@ -72,11 +80,20 @@ class RIRecord:
 class RegistryHive:
     CONTROL_SETS = [r'\ControlSet001', r'\ControlSet002']
 
-    def __init__(self, hive_path):
+    def __init__(self, hive_path, hive_type=None, partial_hive_path=None):
         """
         Represents a registry hive
         :param hive_path: Path to the registry hive
+        :param hive_type: The hive type can be specified if this is a partial hive,
+                          or for some other reason regipy cannot identify the hive type
+        :param partial_hive_path: The path from which the partial hive actually starts, for example:
+                                  hive_type=ntuser partial_hive_path="/Software" would mean
+                                  this is actually a HKCU hive, starting from HKCU/Software
         """
+
+        self.partial_hive_path = None
+        self.hive_type = None
+
         with open(hive_path, 'rb') as f:
             self._stream = BytesIO(f.read())
 
@@ -89,10 +106,20 @@ class RegistryHive:
             self.root = NKRecord(root_hbin_cell, s)
         self.name = self.header.file_name
 
-        try:
-            self.hive_type = identify_hive_type(self.name)
-        except UnidentifiedHiveException:
-            self.hive_type = None
+        if hive_type:
+            if hive_type.lower() in SUPPORTED_HIVE_TYPES:
+                self.hive_type = hive_type
+            else:
+                raise UnidentifiedHiveException(f'{hive_type} is not a supported hive type: '
+                                                f'only the following are supported: {SUPPORTED_HIVE_TYPES}')
+        else:
+            try:
+                self.hive_type = identify_hive_type(self.name)
+            except UnidentifiedHiveException:
+                logger.info(f'Hive type for {hive_path} was not identified: {self.name}')
+
+        if partial_hive_path:
+            self.partial_hive_path = partial_hive_path
 
     def recurse_subkeys(self, nk_record=None, path=None, as_json=False):
         """
@@ -106,24 +133,34 @@ class RegistryHive:
             nk_record = self.root
 
         # Iterate over subkeys
-        for subkey in nk_record.iter_subkeys():
-            if subkey.subkey_count:
+        if nk_record.header.subkey_count:
+            for subkey in nk_record.iter_subkeys():
                 if path:
-                    yield from self.recurse_subkeys(nk_record=subkey, path=r'{}\{}'.format(path, subkey.name),
+                    subkey_path = r'{}\{}'.format(path, subkey.name) if path else r'\{}'.format(subkey.name)
+                else:
+                    subkey_path = f'\\{subkey.name}'
+
+                # Leaf Index records do not contain subkeys
+                if isinstance(subkey, LIRecord):
+                    continue
+
+                if subkey.subkey_count:
+                    yield from self.recurse_subkeys(nk_record=subkey,
+                                                    path=subkey_path,
                                                     as_json=as_json)
-                else:
-                    yield from self.recurse_subkeys(nk_record=subkey, path=r'\{}'.format(subkey.name), as_json=as_json)
 
-            values = []
-            if subkey.values_count:
-                if as_json:
-                    values = [attr.asdict(x) for x in subkey.iter_values(as_json=as_json)]
-                else:
-                    values = list(subkey.iter_values(as_json=as_json))
+                values = []
+                if subkey.values_count:
+                    if as_json:
+                        values = [attr.asdict(x) for x in subkey.iter_values(as_json=as_json)]
+                    else:
+                        values = list(subkey.iter_values(as_json=as_json))
 
-            yield Subkey(subkey_name=subkey.name, path=r'{}\{}'.format(path, subkey.name) if path else '\\',
-                         timestamp=convert_wintime(subkey.header.last_modified), values=values,
-                         values_count=len(values))
+                ts = convert_wintime(subkey.header.last_modified)
+                yield Subkey(subkey_name=subkey.name, path=subkey_path,
+                             timestamp=ts.isoformat() if as_json else ts, values=values,
+                             values_count=len(values),
+                             actual_path=f'{self.partial_hive_path}{subkey_path}' if self.partial_hive_path else None)
 
         # Get the values of the subkey
         values = []
@@ -133,8 +170,11 @@ class RegistryHive:
             else:
                 values = list(nk_record.iter_values(as_json=as_json))
 
-        yield Subkey(subkey_name=nk_record.name, path=path,
-                     timestamp=convert_wintime(nk_record.header.last_modified), values=values, values_count=len(values))
+        ts = convert_wintime(nk_record.header.last_modified)
+        subkey_path = path or '\\'
+        yield Subkey(subkey_name=nk_record.name, path=subkey_path,
+                     timestamp=ts.isoformat() if as_json else ts, values=values, values_count=len(values),
+                     actual_path=f'{self.partial_hive_path}\\{subkey_path}' if self.partial_hive_path else None)
 
     def get_hbin_at_offset(self, offset=0):
         """
@@ -146,18 +186,24 @@ class RegistryHive:
         return HBin(self._stream)
 
     def get_key(self, key_path):
+        if self.partial_hive_path:
+            if key_path.startswith(self.partial_hive_path):
+                key_path = key_path.partition(self.partial_hive_path)[-1]
+            else:
+                raise RegistryKeyNotFoundException(f'Did not find subkey at {key_path}, because this is a partial hive')
+
         logger.debug('Getting key: {}'.format(key_path))
 
         if key_path == '\\':
             return self.root
 
         key_path_parts = key_path.split('\\')[1:]
-        previous_key_name = ['root']
+        previous_key_name = []
+
         subkey = self.root.get_key(key_path_parts.pop(0))
 
         if not subkey:
-            raise RegistryKeyNotFoundException(
-                'Did not find {} at {}'.format(key_path, r'\\'.join(previous_key_name)))
+            raise RegistryKeyNotFoundException('Did not find subkey at {}'.format(key_path))
 
         if not key_path_parts:
             return subkey
@@ -188,6 +234,11 @@ class RegistryHive:
         logger.info('Found control sets: {}'.format(result))
         return result
 
+    def dump_hive_to_json(self, output_path, name_key_entry, verbose=False):
+        with jsonlines.open(output_path, mode='w') as writer:
+            for entry in tqdm(self.recurse_subkeys(name_key_entry, as_json=True), disable=not verbose):
+                writer.write(attr.asdict(entry))
+
 
 class HBin:
     def __init__(self, stream):
@@ -254,39 +305,50 @@ class NKRecord:
     def iter_subkeys(self):
 
         if not self.header.subkey_count:
-            raise NoRegistrySubkeysException('No subkeys for {}'.format(self.name))
+            return None
 
         # Go to the offset where the subkey list starts (+4 is because of the cell header)
         target_offset = REGF_HEADER_SIZE + 4 + self.header.subkeys_list_offset
         self._stream.seek(target_offset)
 
         # Read the signature
-        signature = Bytes(2).parse_stream(self._stream)
+        try:
+            signature = Bytes(2).parse_stream(self._stream)
+        except StreamError as ex:
+            raise RegistryParsingException(f'Bad subkey at offset {target_offset}: {ex}')
 
-        # LF and LH contain subkeys
-        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE]:
-            yield from self._parse_subkeys(self._stream)
-        elif signature == LEAF_INDEX_SIGNATURE:
-            yield LIRecord(Int32ul.parse_stream(self._stream))
+        # LF,  LH and RI contain subkeys
+        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE, LEAF_INDEX_SIGNATURE]:
+            yield from self._parse_subkeys(self._stream, signature=signature)
         # RI contains pointers to arrays of subkeys
         elif signature == INDEX_ROOT_SIGNATURE:
             ri_record = RIRecord(self._stream)
             if ri_record.header.element_count > 0:
                 for element in ri_record.header.elements:
                     # We skip 6 because of the signature as well as the cell header
-                    self._stream.seek(REGF_HEADER_SIZE + 6 + element.subkey_list_offset)
+                    element_target_offset = REGF_HEADER_SIZE + 4 + element.subkey_list_offset
+                    self._stream.seek(element_target_offset)
                     yield from self._parse_subkeys(self._stream)
 
     @staticmethod
-    def _parse_subkeys(stream):
+    def _parse_subkeys(stream, signature=None):
         """
-        Parse an LF or LH Record
+        Parse an LI , LF or LH Record
         :param stream: A stream at the header of the LH or LF entry, skipping the signature
         :return:
         """
-        subkeys = LF_LH_SK_ELEMENT.parse_stream(stream)
+        if not signature:
+            signature = stream.read(2)
+
+        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE]:
+            subkeys = LF_LH_SK_ELEMENT.parse_stream(stream)
+        elif signature == LEAF_INDEX_SIGNATURE:
+            subkeys = INDEX_LEAF.parse_stream(stream)
+        else:
+            raise RegistryParsingException(f'Expected a known signature, got: {signature} at offset {stream.tell()}')
+
         for subkey in subkeys.elements:
-            stream.seek(REGF_HEADER_SIZE + subkey.named_key_offset)
+            stream.seek(REGF_HEADER_SIZE + subkey.key_node_offset)
 
             # This cell should always be allocated, therefor we expect a negative size
             cell_size = Int32sl.parse_stream(stream) * -1
@@ -339,7 +401,7 @@ class NKRecord:
         :return: List of values for the subkey
         """
         if not self.values_count:
-            raise NoRegistryValuesException(f'No registry values for {self.name}')
+            return
 
         # Get the offset of the values key. We skip 4 because of Cell Header
         target_offset = REGF_HEADER_SIZE + 4 + self.header.values_list_offset
@@ -351,7 +413,12 @@ class NKRecord:
             with boomerang_stream(self._stream) as substream:
                 actual_vk_offset = REGF_HEADER_SIZE + 4 + vk_offset
                 substream.seek(actual_vk_offset)
-                vk = VALUE_KEY.parse_stream(substream)
+                try:
+                    vk = VALUE_KEY.parse_stream(substream)
+                except ConstError:
+                    logger.error(f'Could not parse VK at {substream.tell()}, registry hive is probably corrupted.')
+                    return
+
                 value = self.read_value(vk, substream)
 
                 if vk.name_size == 0:
@@ -359,13 +426,29 @@ class NKRecord:
                 else:
                     value_name = vk.name.decode(errors='replace')
 
+                # If the value is bigger than this value, it means this is a DEVPROP structure
+                # https://doxygen.reactos.org/d0/dba/devpropdef_8h_source.html
+                # https://sourceforge.net/p/mingw-w64/mingw-w64/ci/668a1d3e85042c409e0c292e621b3dc0aa26177c/tree/
+                # mingw-w64-headers/include/devpropdef.h?diff=dd86a3b7594dadeef9d6a37c4b6be3ca42ef7e94
+                # We currently do not support these, but also wouldn't like to yield this as binary data
+                # This int casting will always work because the data_type is construct's EnumIntegerString
+                # TODO: Add actual parsing
+                if int(vk.data_type) > 0xffff0000:
+                    data_type = int(vk.data_type) & 0xffff
+                    continue
+
+                # Skip this unknown data type, research pending :)
+                # TODO: Add actual parsing
+                if int(vk.data_type) == 0x200000:
+                    continue
+
                 data_type = str(vk.data_type)
                 if data_type in ['REG_SZ', 'REG_EXPAND', 'REG_EXPAND_SZ']:
                     if vk.data_size >= 0x80000000:
                         # data is contained in the data_offset field
                         value.size -= 0x80000000
                         actual_value = vk.data_offset
-                    elif vk.data_size > 0x3fd8:
+                    elif vk.data_size > 0x3fd8 and value.value[:2] == b'db':
                         data = self._parse_indirect_block(substream, value)
                         actual_value = try_decode_binary(data, as_json=as_json)
                     else:
@@ -374,9 +457,10 @@ class NKRecord:
                     if vk.data_size >= 0x80000000:
                         # data is contained in the data_offset field
                         actual_value = vk.data_offset
-                    elif vk.data_size > 0x3fd8:
+                    elif vk.data_size > 0x3fd8 and value.value[:2] == b'db':
                         try:
                             actual_value = self._parse_indirect_block(substream, value)
+
                             actual_value = try_decode_binary(actual_value, as_json=True) if as_json else actual_value
                         except ConstError:
                             logger.error(f'Bad value at {actual_vk_offset}')
@@ -395,15 +479,20 @@ class NKRecord:
                     parsed_value = GreedyRange(CString('utf-16-le')).parse(value.value)
                     # Because the ListContainer object returned by Construct cannot be turned into a list,
                     # we do this trick
-                    actual_value = str(parsed_value) if as_json else [x for x in parsed_value if x]
+                    actual_value = [x for x in parsed_value if x]
+                # We currently dumps this as hex string or raw
+                # TODO: Add actual parsing
+                elif data_type in ['REG_RESOURCE_REQUIREMENTS_LIST', 'REG_RESOURCE_LIST']:
+                    actual_value = binascii.b2a_hex(value.value).decode()[:max_len] if as_json else value.value
                 else:
                     actual_value = try_decode_binary(value.value, as_json=as_json)
                 yield Value(name=value_name, value_type=str(value.value_type), value=actual_value,
                             is_corrupted=is_corrupted)
 
-    def get_value(self, value_name, as_json=False, raise_on_missing=False):
+    def get_value(self, value_name=DEFAULT_VALUE, as_json=False, raise_on_missing=False):
         """
-        Get a value by name. Will raise if not found, except a default is given
+        Get a value by name. Will raise if raise_on_missing is set,
+        if no value name is given, will return the content of the default value
         :param value_name: The value name to look for
         :param as_json: Whether to normalize the data as JSON or not
         :param raise_on_missing: Will raise exception if value is missing, else will return None
@@ -417,6 +506,9 @@ class NKRecord:
             raise RegistryValueNotFoundException('Did not find the value {} on subkey {}'.format(value_name, self.name))
         return None
 
+    def get_values(self, as_json=False):
+        return [x for x in self.iter_values(as_json=as_json)]
+
     def __dict__(self):
         return {
             'name': self.name,
